/**
 * Stub for GtkHotkey.Win32Listener.
 */
using Gdk;

namespace GtkHotkey
{
  public class Win32Hotkey : Object
  {
    public Win32.Atom atom { get; construct; }
    public Info info { get; construct; }
    public Win32Hotkey (Win32.Atom atom, Info info)
    {
      this.atom = atom;
      this.info = info;
    }
    ~Win32HotKey ()
    {
      this.atom.delete_global ();
    }
  }

  public class Win32Listener : Listener
  {

    private Gdk.Window hidden = null;
    private Win32.WindowHandle hwnd;
    private HashTable<Win32.Atom?,Win32Hotkey> hotkey_registry;

    construct
    {
      this.hidden = this.create_hidden_msg_window ();
      this.hwnd = Win32.WindowHandle.from_gdk_window (this.hidden);
      this.hotkey_registry = new HashTable<Win32.Atom?,Win32Hotkey> (int_hash, int_equal);
      this.hidden.add_filter (this.hotkey_filter);
    }

    ~Win32Listener ()
    {
      List<weak Win32.Atom?>? keys = this.hotkey_registry.get_keys ();
      foreach (weak Win32.Atom atom in keys)
      {
        this.do_remove_hotkey (atom);
      }
      this.hidden.remove_filter (this.hotkey_filter); // XXX Fix VAPI
      this.hidden.destroy ();
      this.hidden = null;
    }

    private void
    do_remove_hotkey (Win32.Atom atom)
    {
      Win32.HotKey.unregister (this.hwnd, atom);
      this.hotkey_registry.remove (atom);
    }

    /**
     * "Ported" from GTK+ SVN revision 20900, file
     * </modules/engines/ms-windows/msw_theme_main.c>.
     * Licensed under the LGPL-2 or later.
     */
    private Window
    create_hidden_msg_window ()
    {
      WindowAttr attributes = WindowAttr ();
      int attributes_mask;

      attributes.x = -100;
      attributes.y = -100;
      attributes.width = 10;
      attributes.height = 10;
      attributes.window_type = WindowType.TEMP;
      attributes.wclass = WindowClass.INPUT_ONLY;
      attributes.override_redirect = true;
      attributes.event_mask = 0;

      attributes_mask = WindowAttributesType.X |
                        WindowAttributesType.Y |
                        WindowAttributesType.NOREDIR;
      return new Window (get_default_root_window (), attributes,
                         attributes_mask);
    }

    private Win32.Atom atomize (Info hotkey)
    {
      string key;
      Win32.Atom atom;
      key = hotkey.application_id + ":" + hotkey.key_id;
      atom = Win32.Atom.find_global (key);
      if (atom == 0)
      {
        return Win32.Atom.add_global (key);
      }
      else
      {
        return atom;
      }
    }

    private FilterReturn
    hotkey_filter (XEvent xevt, Event evt)
    {
      weak Win32.Message* p_msg = (Win32.Message*)xevt;
      weak Win32.Message msg = *p_msg;
      if (msg.message == Win32.Message.HOTKEY)
      {
        Win32Hotkey? wh = this.hotkey_registry.lookup ((Win32.Atom)msg.wParam);
        if (wh != null)
        {
          this.activated (wh.info);
          return Gdk.FilterReturn.REMOVE;
        }
      }
      return Gdk.FilterReturn.CONTINUE;
    }

    private uint
    convert_modifier_mask (Gdk.ModifierType modifier)
    {
      uint win32_modifier = 0;
      if ((modifier & ModifierType.SHIFT_MASK) != 0)
      {
        win32_modifier |= Win32.HotKey.MOD_SHIFT;
      }
      if ((modifier & ModifierType.CONTROL_MASK) != 0)
      {
        win32_modifier |= Win32.HotKey.MOD_CONTROL;
      }
      if ((modifier & ModifierType.MOD1_MASK) != 0)
      {
        win32_modifier |= Win32.HotKey.MOD_ALT;
      }
      if ((modifier & ModifierType.SUPER_MASK) != 0 ||
          (modifier & ModifierType.HYPER_MASK) != 0)
      {
        win32_modifier |= Win32.HotKey.MOD_WIN;
      }
      return win32_modifier;
    }

    private uint
    convert_keyval (uint keyval)
    {
      weak Gdk.Keymap km;
      weak Gdk.KeymapKey[]? keys;
      km = Gdk.Keymap.get_default ();
      if (km.get_entries_for_keyval (keyval, out keys))
      {
        return keys[0].keycode;
      }
      else
      {
        return 0;
      }
    }

    public override void bind_hotkey (Info hotkey)
    {
      weak Win32Hotkey? wh = null;
      Win32.Atom atom = this.atomize (hotkey);
      wh = this.hotkey_registry.lookup (atom);
      if (wh == null)
      {
        bool result;
        uint keyval, modifier, keycode;
        Gdk.ModifierType gdk_modifier;
        Gtk.accelerator_parse (hotkey.signature, out keyval,
                               out gdk_modifier);
        modifier = this.convert_modifier_mask (gdk_modifier);
        keycode = this.convert_keyval (keyval);
        result = Win32.HotKey.register (this.hwnd, atom, modifier, keycode);
        if (result)
        {
          this.hotkey_registry.insert (atom, new Win32Hotkey (atom, hotkey));
        }
      }
    }
    public override void unbind_hotkey (Info hotkey)
    {
      weak Win32Hotkey? wh = null;
      Win32.Atom atom = this.atomize (hotkey);
      wh = this.hotkey_registry.lookup (atom);
      if (wh != null)
      {
        this.do_remove_hotkey (atom);
      }
    }
  }
}
/* vim: set ts=2 sts=2 sw=2 et ai cindent : */
